@@ -19,12 +19,16 @@
 
 #include <QAction>
 #include <QApplication>
-#include <QFontDatabase>
-#include <QIcon>
 #include <QClipboard>
+#include <QDir>
+#include <QFontDatabase>
 #include <QGridLayout>
+#include <QIcon>
+#include <QMenu>
 
 #include <KActionCollection>
+#include <KBookmarkManager>
+#include <KBookmarkMenu>
 #include <KConfigGroup>
 #include <KLocalizedString>
 #include <KStandardAction>
@@ -32,6 +36,51 @@
 #include <KSharedConfig>
 #include <KToggleAction>
 
+class KCharSelectBookmarkOwner : public KBookmarkOwner
+{
+public:
+    KCharSelectBookmarkOwner(KCharSelectDia *dia)
+        : d(dia) { }
+    ~KCharSelectBookmarkOwner() { }
+
+    QString currentTitle() const Q_DECL_OVERRIDE
+    {
+        uint ucs4 = d->charSelect->currentCodePoint();
+        if (QChar::isPrint(ucs4)) {
+            return QString::fromUcs4(&ucs4, 1) + QLatin1Char(' ') + formatCodePoint(ucs4);
+        } else {
+            return formatCodePoint(ucs4);
+        }
+    }
+
+    QUrl currentUrl() const Q_DECL_OVERRIDE
+    {
+        return QUrl(formatCodePoint(d->charSelect->currentCodePoint()));
+    }
+
+    void openBookmark(const KBookmark &bm, Qt::MouseButtons, Qt::KeyboardModifiers) Q_DECL_OVERRIDE
+    {
+        QString c = bm.url().toString(QUrl::PreferLocalFile | QUrl::RemoveScheme);
+        if (c.startsWith(QLatin1String("U+"))) {
+            uint uc = c.mid(2).toUInt(Q_NULLPTR, 16);
+            d->charSelect->setCurrentCodePoint(uc);
+        }
+    }
+
+private:
+    static QString formatCodePoint(uint ucs4)
+    {
+        QString c = QString::number(ucs4, 16).toUpper();
+        while (c.length() < 4) {
+            c.prepend(QLatin1Char('0'));
+        }
+        return QStringLiteral("U+") + c;
+    }
+
+private:
+    KCharSelectDia *d;
+};
+
 /******************************************************************/
 /* class KCharSelectDia                                           */
 /******************************************************************/
@@ -44,7 +93,7 @@ KCharSelectDia::KCharSelectDia()
   KConfigGroup gr = config->group("General");
 
   vFont = gr.readEntry("selectedFont", QFontDatabase::systemFont(QFontDatabase::GeneralFont));
-  vChr = QChar(static_cast<unsigned short>(gr.readEntry("char", 33)));
+  vChr = gr.readEntry("char", 33);
   _rtl = gr.readEntry("rtl", false);
 
   QWidget *mainWidget = new QWidget(this);
@@ -54,12 +103,13 @@ KCharSelectDia::KCharSelectDia()
 
   // Add character selection widget from library kdeui
   charSelect = new KCharSelect(mainWidget, actionCollection());
-  charSelect->setCurrentChar(vChr);
+  charSelect->setAllPlanesEnabled(true);
+  charSelect->setCurrentCodePoint(vChr);
   charSelect->setCurrentFont(vFont);
   charSelect->resize(charSelect->sizeHint());
-  connect(charSelect, &KCharSelect::currentCharChanged, this, &KCharSelectDia::charChanged);
-  connect(charSelect,SIGNAL(charSelected(QChar)),
-         SLOT(add(QChar)));
+  connect(charSelect, &KCharSelect::currentCodePointChanged, this, &KCharSelectDia::charChanged);
+  connect(charSelect,SIGNAL(codePointSelected(uint)),
+         SLOT(add(uint)));
 
   connect(charSelect, &KCharSelect::currentFontChanged, this, &KCharSelectDia::fontSelected);
   grid->addWidget(charSelect, 0, 0, 1, 4);
@@ -123,6 +173,21 @@ KCharSelectDia::KCharSelectDia()
   else
     lined->setAlignment( Qt::AlignLeft );
 
+  QString filename = QStandardPaths::locate(QStandardPaths::GenericDataLocation, QStringLiteral("kcharselect/bookmarks.xml"));
+  if (filename.isEmpty()) {
+    filename = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QStringLiteral("/kcharselect");
+    QDir().mkpath(filename);
+    filename += QStringLiteral("/bookmarks.xml");
+  }
+  KBookmarkManager* manager = KBookmarkManager::managerForFile(filename, QStringLiteral("kcharselect"));
+
+  action = actionCollection()->addAction(QStringLiteral("bookmarks"));
+  action->setText(i18n("Bookmarks"));
+  QMenu *bmmenu = new QMenu(this);
+  action->setMenu(bmmenu);
+  KBookmarkMenu *bm = new KBookmarkMenu(manager, new KCharSelectBookmarkOwner(this), bmmenu, actionCollection());
+  bm->setParent(this);
+
   setupGUI(ToolBar|Keys|Save|Create);
 }
 
@@ -133,14 +198,14 @@ void KCharSelectDia::closeEvent(QCloseEvent *event)
   KConfigGroup gr = config->group("General");
 
   gr.writeEntry("selectedFont", vFont);
-  gr.writeEntry("char", static_cast<int>(vChr.unicode()));
+  gr.writeEntry("char", vChr);
   gr.writeEntry("rtl", _rtl);
 
   KXmlGuiWindow::closeEvent(event);
 }
 
 //==================================================================
-void KCharSelectDia::charChanged(const QChar &_chr)
+void KCharSelectDia::charChanged(uint _chr)
 {
   vChr = _chr;
 }
@@ -154,15 +219,15 @@ void KCharSelectDia::fontSelected(const QFont &_font)
 }
 
 //==================================================================
-void KCharSelectDia::add(const QChar &_chr)
+void KCharSelectDia::add(uint _chr)
 {
   charChanged(_chr);
 
   QString str = lined->text();
   const int pos = lined->cursorPosition();
-  str.insert(pos, _chr);
+  str.insert(pos, QString::fromUcs4(&_chr, 1));
   lined->setText(str);
-  lined->setCursorPosition(pos + 1);
+  lined->setCursorPosition(pos + QString::fromUcs4(&_chr, 1).size());
 }
 
 //==================================================================
@@ -170,7 +235,7 @@ void KCharSelectDia::toClip()
 {
   QString str = lined->text();
   if (str.isEmpty())
-    str = vChr;
+    str = QString::fromUcs4(&vChr, 1);
   QClipboard *cb = QApplication::clipboard();
   cb->setText(str,QClipboard::Clipboard);
   cb->setText(str,QClipboard::Selection);
@@ -185,7 +250,7 @@ void KCharSelectDia::toClipUTF8()
   QClipboard *cb = QApplication::clipboard();
   QString str = lined->text();
   if (str.isEmpty())
-    str = vChr;
+    str = QString::fromUcs4(&vChr, 1);
   cb->setText(QLatin1String( str.toUtf8() ));
 }
 
@@ -206,7 +271,7 @@ void KCharSelectDia::toClipHTML()
 
   input = lined->text();
   if (input.isEmpty())
-    input = vChr;
+    input = QString::fromUcs4(&vChr, 1);
   const int inputLength = input.length();
   for(i=0; i< inputLength; ++i )
     {
